import click


class ManagePlugin(object):

    name = 'manage'

    def __init__(self):
        self.cli = click.group()(lambda: None)

    def setup(self, app):

        self.app = app

        app.config.setdefault('MANAGE_SHELL', {'app': app})

        @self.command()
        def shell():
            """ Run the application shell. """
            from werkzeug import script

            ctx = app.config['MANAGE_SHELL']
            if callable(ctx):
                ctx = ctx()

            return script.make_shell(lambda: ctx, "Loaded objects: " + ", ".join(ctx.keys()))()

        @self.command()
        # @self.option('--reload/--no-reload', default=False)
        # @self.option('--debug/--no-debug', default=False)
        @self.option('--port', default=5000)
        def runserver(**kwargs):
            """ Run the application. """
            app.run(**kwargs)

    @staticmethod
    def option(*args, **kwargs):
        return click.option(*args, **kwargs)

    def command(self, *args, **kwargs):
        return self.cli.command(*args, **kwargs)

    def shell(self, func):
        self.app.config['MANAGE_SHELL'] = func

    def __call__(self, *args, **kwargs):
        return self.cli(*args, **kwargs) # noqa

# pylama:ignore=E1103,W0612



# Configure your tests here
import asyncio

import pytest
import webtest
from aiohttp.protocol import HttpVersion11, HttpVersion10, RawRequestMessage
from aiohttp.multidict import CIMultiDict
from aiohttp.web import (
    AbstractMatchInfo,
    HTTPException,
    Request,
    RequestHandler,
    RequestHandlerFactory,
    StreamResponse,
)


class TestRequestHandler(RequestHandler):

    @asyncio.coroutine
    def handle_request(self, message, payload):
        app = self._app
        request = Request(app, message, payload, self.transport, self.reader, self.writer)
        try:
            match_info = yield from self._router.resolve(request)

            assert isinstance(match_info, AbstractMatchInfo), match_info

            request._match_info = match_info
            handler = match_info.handler

            for factory in reversed(self._middlewares):
                handler = yield from factory(app, handler)
            resp = yield from handler(request)

            if not isinstance(resp, StreamResponse):
                raise RuntimeError(
                    ("Handler {!r} should return response instance, "
                     "got {!r} [middlewares {!r}]").format(
                         match_info.handler,
                         type(resp),
                         self._middlewares))
        except HTTPException as exc:
            resp = exc

        return resp


class TestRequest(webtest.TestRequest):

    """ Support asyncio loop. """

    def call_application(self, application, catch_exc_info=False):
        if self.is_body_seekable:
            self.body_file_raw.seek(0)

        http_version = HttpVersion10 if self.http_version == 'HTTP/1.0' else HttpVersion11
        message = RawRequestMessage(
            self.method, self.path, http_version, CIMultiDict(self.headers), False, False)
        payload = self.body_file_raw

        loop = asyncio.get_event_loop()
        handler = RequestHandlerFactory(
            application, application.router, handler=TestRequestHandler, loop=loop)()
        response = loop.run_until_complete(handler.handle_request(message, payload))

        return response.status, response.headers.items(), [response.body], None


class TestApp(webtest.TestApp):

    RequestClass = TestRequest


@pytest.fixture(scope='session')
def app():
    """ Provide an example application. """
    from app import app
    return app


@pytest.fixture(scope='session')
def loop(request):
    """ Create and provide asyncio loop. """
    loop = asyncio.new_event_loop()
    request.addfinalizer(lambda: loop.close())
    return loop


@pytest.fixture(scope='function')
def client(app, loop):
    client = TestApp(app, lint=False)
    client.exception = webtest.AppError
    return client



# Package information
# ===================
import asyncio
import concurrent
import datetime
from functools import partial
import threading

import peewee
from playhouse.shortcuts import model_to_dict, dict_to_model
from playhouse.db_url import connect


class AsyncDatabaseMixin:

    max_connections = 4

    def __init__(self, database, loop=None, **kwargs):
        self.connections = {}
        super().__init__(database, **kwargs)
        self.loop = loop or asyncio.get_event_loop()
        self.threadpool = concurrent.futures.ThreadPoolExecutor(max_workers=self.max_connections)

    @property
    def _Database__local(self):
        tid = threading.get_ident()
        return self.connections.get(tid, peewee._ConnectionLocal())

    @_Database__local.setter
    def _Database__local(self, value):
        tid = threading.get_ident()
        self.connections[tid] = value


class PeeweePlugin(object):

    """ Integrate peewee to bottle. """

    name = 'peewee'
    default_connection = 'sqlite:///db.sqlite'
    max_connections = 2

    def __init__(self, connection=None):
        self.app = None
        self.database = peewee.Proxy()
        self.connection = connection or self.default_connection
        self.serializer = Serializer()
        self.threadpool = concurrent.futures.ThreadPoolExecutor(max_workers=self.max_connections)

    def setup(self, app):
        """ Initialize the application. """

        self.app = app
        app.config.setdefault('PEEWEE_CONNECTION', self.connection)

        self.connection = app.config.get('PEEWEE_CONNECTION')
        self.database.initialize(connect(self.connection))

    def query(self, query):
        if isinstance(query, peewee.SelectQuery):
            return self.run(lambda: list(query))
        return self.run(query.execute)

    @asyncio.coroutine
    def run(self, function, *args, **kwargs):
        """ Run sync code asyncronously. """

        if kwargs:
            function = partial(function, **kwargs)

        def iteration(database, *args):
            database.connect()
            try:
                with database.transaction():
                    return function(*args)
            except peewee.PeeweeException:
                database.rollback()
                raise
            finally:
                database.commit()
                if not database.is_closed():
                    database.close()

        result = yield from self.app.loop.run_in_executor(
            self.threadpool, iteration, self.database,  *args)
        return result

    def to_dict(self, obj, **kwargs):
        return self.serializer.serialize_object(obj, **kwargs)


class Serializer(object):
    date_format = '%Y-%m-%d'
    time_format = '%H:%M:%S'
    datetime_format = ' '.join([date_format, time_format])

    def convert_value(self, value):
        if isinstance(value, datetime.datetime):
            return value.strftime(self.datetime_format)

        if isinstance(value, datetime.date):
            return value.strftime(self.date_format)

        if isinstance(value, datetime.time):
            return value.strftime(self.time_format)

        if isinstance(value, peewee.Model):
            return value.get_id()

        return value

    def clean_data(self, data):
        for key, value in data.items():
            if isinstance(value, dict):
                self.clean_data(value)
            elif isinstance(value, (list, tuple)):
                data[key] = map(self.clean_data, value)
            else:
                data[key] = self.convert_value(value)
        return data

    def serialize_object(self, obj, **kwargs):
        data = model_to_dict(obj, **kwargs)
        return self.clean_data(data)


class Deserializer(object):

    @staticmethod
    def deserialize_object(model, data):
        return dict_to_model(model, data)



import asyncio
import json
import signal
import logging
import os
from importlib import import_module

from aiohttp import web
from cached_property import cached_property


CONFIGURATION_ENVIRON_VARIABLE = 'MUFFIN_CONFIG'


class Application(web.Application):

    """ Do some stuff. """

    # Default application settings
    __defaults = {

        # Install the plugins
        'PLUGINS': (
            'muffin.plugins.manage:ManagePlugin',
            'muffin.plugins.jade:JadePlugin',
            'muffin.plugins.peewee:PeeweePlugin',
        ),

        # Enable debug mode
        'DEBUG': False
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.loop.set_debug(self.config['DEBUG'])
        self._middlewares = list(self._middlewares)
        self.plugins = []
        for plugin in self.config['PLUGINS']:
            self.install(plugin)

    @cached_property
    def config(self):
        """ Load the application configuration. """
        config = dict(self.__defaults)
        module = os.environ.get(CONFIGURATION_ENVIRON_VARIABLE, 'config')
        try:
            module = import_module(module)
            config.update({
                name: getattr(module, name) for name in dir(module)
                if name == name.upper() and not name.startswith('_')
            })
        except ImportError:
            self.logger.warn("The configuration hasn't found: %s" % module)

        return config

    def install(self, plugin):
        """ Install plugin to the application. """

        if isinstance(plugin, str):
            module, attr = plugin.split(':')
            module = import_module(module)
            plugin = getattr(module, attr)

        if isinstance(plugin, type):
            plugin = plugin()

        if hasattr(plugin, 'setup'):
            plugin.setup(self)

        if hasattr(plugin, 'middleware_factory'):
            self.middlewares.append(getattr(plugin.middleware))

        self.plugins.append(plugin.name)
        setattr(self, plugin.name, plugin)

    def run(self, host='127.0.0.1', port=8080):
        """ Run the application. """

        ch = logging.StreamHandler()
        ch.setFormatter(logging.Formatter('%(asctime)-15s - %(message)s'))
        self.logger.addHandler(ch)
        self.logger.setLevel('DEBUG') if self.config['DEBUG'] else self.logger.setLevel('WARNING')
        self.loop.run_until_complete(
            self.loop.create_server(self.make_handler(), host, port))

        self.loop.add_signal_handler(signal.SIGTERM, self.stop)
        self.loop.add_signal_handler(signal.SIGINT, self.stop)

        try:
            self.logger.info('Server started at http://%s:%s' % (host, port))
            self.loop.run_forever()
            asyncio.sleep
        finally:
            self.loop.close()

    def stop(self):
        self.logger.info('The server is stopping.')
        self.loop.stop()

    def view(self, path, method='GET', name=None):
        """ Convert a view to couroutine and bind a route. """
        def wrapper(view):
            if not asyncio.iscoroutinefunction(view):
                view = asyncio.coroutine(view)

            def wrap_response(request, **kwargs):
                response = yield from view(request, **kwargs)

                if isinstance(response, (list, dict)):
                    response = web.Response(text=json.dumps(response),
                                            content_type='application/json')

                elif not isinstance(response, web.Response):
                    response = web.Response(text=str(response), content_type='text/html')

                self.logger.info('%s %d %s', request.method, response.status, request.path)
                return response

            self.router.add_route(method, path, wrap_response, name=name)

        return wrapper
